/***************************************************************************
 * Copyright 2013 Kieker Project (http://kieker-monitoring.net)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ***************************************************************************/

package kieker.webgui.service.impl;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

import kieker.webgui.common.exception.GraphLayoutException;
import kieker.webgui.service.IGraphLayoutService;

import org.eclipse.emf.common.util.EList;

import de.cau.cs.kieler.core.alg.BasicProgressMonitor;
import de.cau.cs.kieler.core.alg.IKielerProgressMonitor;
import de.cau.cs.kieler.core.kgraph.KEdge;
import de.cau.cs.kieler.core.kgraph.KNode;
import de.cau.cs.kieler.core.kgraph.KPort;
import de.cau.cs.kieler.kiml.AbstractLayoutProvider;
import de.cau.cs.kieler.kiml.klayoutdata.KEdgeLayout;
import de.cau.cs.kieler.kiml.klayoutdata.KPoint;
import de.cau.cs.kieler.kiml.klayoutdata.KShapeLayout;
import de.cau.cs.kieler.kiml.options.EdgeRouting;
import de.cau.cs.kieler.kiml.options.LayoutOptions;
import de.cau.cs.kieler.kiml.options.PortConstraints;
import de.cau.cs.kieler.kiml.util.KimlUtil;
import de.cau.cs.kieler.klay.layered.LayeredLayoutProvider;

import org.springframework.stereotype.Service;

/**
 * This is an implementation of the {@link IGraphLayoutService}. It uses the Kieler layout algorithms to perform the auto layout.
 * 
 * @author Robin Weiss, Nils Christian Ehmke
 */
@Service
public final class KielerGraphLayoutServiceImpl implements IGraphLayoutService {

	/**
	 * Default constructor.
	 */
	public KielerGraphLayoutServiceImpl() {
		// No code necessary
	}

	@Override
	public String layoutGraph(final String nodes, final String edges) throws GraphLayoutException {
		try {
			// Convert the given nodes and edges to a graph
			final LayoutInformation layoutInformation = new LayoutInformation(nodes, edges);

			// Perform the auto layout on the given graph.
			final IKielerProgressMonitor progressMonitor = new BasicProgressMonitor();
			final AbstractLayoutProvider layoutProvider = new LayeredLayoutProvider();
			layoutProvider.doLayout(layoutInformation.getGraph(), progressMonitor);

			// Convert the layouted graph back into a string representation
			final String layoutedGraph = layoutInformation.getPositions();
			final String bendPoints = layoutInformation.getBendPoints();

			final StringBuilder sb = new StringBuilder();
			sb.append("autoLayout").append("#").append(layoutedGraph).append("#").append(bendPoints);
			return sb.toString();

		} catch (final NumberFormatException ex) {
			throw new GraphLayoutException("Invalid Graph Information String!", ex);
		} catch (final Exception ex) { // NOCS
			throw new GraphLayoutException("Unknown Error!", ex);
		}
	}

	/**
	 * A container for the layout information which will be assembled during the layout process.
	 * 
	 * @author Robin Weiss, Nils Christian Ehmke
	 */
	private static final class LayoutInformation {

		private final KNode graph;
		private final Map<String, List<Object>> children;
		private final Object[][] edges;

		/**
		 * Creates a new instance of this class.
		 * 
		 * @param nodesStr
		 *            The string containing the nodes.
		 * @param edgesStr
		 *            The string containing the edges.
		 * 
		 * @throws GraphLayoutException
		 *             If the given graph is empty or invalid.
		 */
		public LayoutInformation(final String nodesStr, final String edgesStr) throws GraphLayoutException {
			final int firstSemicolon = nodesStr.indexOf(';');

			// if there is no semicolon, we got an empty graph
			if (firstSemicolon < 0) {
				throw new GraphLayoutException("Empty Graph");
			}

			final int scale = Integer.parseInt(nodesStr.substring(0, firstSemicolon));

			final String[] nodeInfo = LayoutInformation.correctEmptyArray(nodesStr.substring(firstSemicolon + 1).split(";"));
			final String[] edgeInfo = LayoutInformation.correctEmptyArray(edgesStr.split(";"));

			this.children = new HashMap<String, List<Object>>();
			this.edges = new Object[edgeInfo.length][5];
			this.graph = KimlUtil.createInitializedNode();

			this.addNodes(scale, nodeInfo);
			this.addEdges(edgeInfo);

			// set graph layout options
			final KShapeLayout layout = this.graph.getData(KShapeLayout.class);
			layout.setProperty(LayoutOptions.EDGE_ROUTING, EdgeRouting.ORTHOGONAL);
		}

		public KNode getGraph() {
			return this.graph;
		}

		/**
		 * Reads the (layouted) positions of a constructed KGraph and writes them as integers to a string, separating each number with a space.
		 * 
		 * @return The constructed string.
		 */
		public String getPositions() {
			final StringBuilder sb = new StringBuilder();

			for (final Entry<String, List<Object>> entry : this.children.entrySet()) {
				final String id = entry.getKey();
				final List<Object> nodeAndPortList = entry.getValue();

				// Extract the position and size of the node
				final KShapeLayout layout = ((KNode) nodeAndPortList.get(0)).getData(KShapeLayout.class);

				final int x = Math.round(layout.getXpos() + (layout.getWidth() / 2));
				final int y;
				// HotFix: wrong nodeFamily position: needs to be half the height lower
				if (nodeAndPortList.size() > 2) {
					y = Math.round(layout.getYpos() + (layout.getHeight()));
				} else {
					y = Math.round(layout.getYpos() + (layout.getHeight() / 2));
				}

				final int width = Math.round(layout.getWidth());
				final int height = Math.round(layout.getHeight());

				// Add the node's information to our string
				sb.append(id).append(' ').append(x).append(' ').append(y).append(' ').append(width).append(' ').append(height).append(';');
			}

			return sb.toString();
		}

		/**
		 * This method iterates through all edges and creates a semicolon separated String containing bend points of all edges.
		 * 
		 * @return The string containing the bend points.
		 */
		public String getBendPoints() {
			final StringBuilder sb = new StringBuilder();

			int e;

			final int len = this.edges.length;
			final int el = 4;

			// Iterate through all edges.
			for (int i = 0; i < len; i++) {
				e = 0;
				final KEdge edge = (KEdge) this.edges[i][e];

				final KEdgeLayout layout = edge.getData(KEdgeLayout.class);
				final EList<KPoint> bendPoints = layout.getBendPoints();

				// add IDs and portIndices
				while (++e < el) {
					sb.append(this.edges[i][e].toString());
					sb.append(" ");
				}
				sb.append(this.edges[i][4].toString());

				// iterate bend points
				final int bl = bendPoints.size();
				KPoint point;

				if (bl > 0) {
					point = bendPoints.get(0);
					sb.append(' ');
					sb.append(Math.round(point.getX()));
					sb.append(' ');
					sb.append(Math.round(point.getY()));

					for (int b = 1; b < bl; b++) {
						sb.append(' ');
						point = bendPoints.get(b);
						sb.append(Math.round(point.getX()));
						sb.append(' ');
						sb.append(Math.round(point.getY()));
					}
				}
				sb.append(';');
			}

			return sb.toString();
		}

		/**
		 * Parses a String array, building a graph with nodes and fixed ports.
		 * 
		 * @param scale
		 *            The scale factor of the graph.
		 * @param nodeInfo
		 *            Information about the nodes: isItCustom?, width, height, parent, portInformation
		 * 
		 * @throws NumberFormatException
		 *             The nodeInfo should mostly be space separated numbers. If it is however invalid, this Exception will be thrown
		 */
		private void addNodes(final int scale, final String[] nodeInfo) throws NumberFormatException {

			// retrieve size modifier
			final int dimHalf = scale;
			final int dim = dimHalf * 2;

			// loop variables //
			KNode node;
			String parentId;
			KNode parent;
			KPort port;
			KShapeLayout layout;

			final int l = nodeInfo.length;

			// nodeID : KNode, ParentID

			// -- 1. Define nodes by the properties, given in the string --
			for (int n = 0; n < l; n++) {

				// i is our property array-pointer
				int i = 0;
				final String[] nodeProps = nodeInfo[n].split(" ");

				// add entry to map
				final List<Object> nodeAndPortList = new ArrayList<Object>();
				this.children.put(nodeProps[i++], nodeAndPortList);
				// create empty node
				node = KimlUtil.createInitializedNode();
				nodeAndPortList.add(node);

				// get parentID
				parentId = nodeProps[i++];
				if ("-1".equals(parentId)) {
					parentId = "";
					node.setParent(this.getGraph());
				}
				nodeAndPortList.add(parentId);

				// get width and height
				final int width = Integer.parseInt(nodeProps[i++]);
				final int height = Integer.parseInt(nodeProps[i++]);

				// get Custom Node tag
				final boolean isNodeFamily = "f".equals(nodeProps[i++]);

				// set width and height of KNode
				layout = node.getData(KShapeLayout.class);
				layout.setWidth(width);
				layout.setHeight(height);

				// if node is a NodeFamily, go on and look for ports
				if (isNodeFamily) {

					// nodeFamily size is fixed
					layout.setProperty(LayoutOptions.PORT_CONSTRAINTS, PortConstraints.FIXED_POS);
					int portX = -dimHalf;

					// add Ports
					for (int portType = 0; portType < 3; portType++) {
						// determine port x-position
						if (portType == 1) {
							portX = width - dimHalf;
						}

						final int portCount = Integer.parseInt(nodeProps[i++]);
						int portY = Integer.parseInt(nodeProps[i++]);

						for (int p = 0; p < portCount; p++) {

							// create port
							port = KimlUtil.createInitializedPort();
							port.setNode(node);
							nodeAndPortList.add(port);

							// set layout for Port
							layout = port.getData(KShapeLayout.class);

							// set position and dimension of port
							layout.setWidth(dim);
							layout.setHeight(dimHalf);
							layout.setYpos(portY);
							layout.setXpos(portX);

							portY += dim;
						}
					}
				}
			}

			// -- 2. resolve stored parentIDs to actual parents
			for (final List<Object> nodeAndPortList : this.children.values()) {
				// get KNode and its parentID
				node = (KNode) nodeAndPortList.get(0);
				parentId = nodeAndPortList.get(1).toString();

				// set other KNode as parent
				if (!parentId.isEmpty()) {
					parent = (KNode) this.children.get(parentId).get(0);
					node.setParent(parent);
				}
			}
		}

		/**
		 * Parses an array which must contain quadruples of numbers.<br/>
		 * The first number must be the index of the children-array, where
		 * the source node is located.<br/>
		 * The second number must be the index of the children-array, where
		 * the target node is located.<br/>
		 * The third number must be the index of the source node List, where
		 * the source port is located.<br/>
		 * The fourth number must be the index of the target node List, where
		 * the target port is located.
		 * 
		 * @param edgeInfo
		 *            The single elements for the edges.
		 * 
		 * @throws NumberFormatException
		 *             If the given graph is invalid.
		 */
		private void addEdges(final String[] edgeInfo) throws NumberFormatException {
			// Set up some loop variables
			String id;
			int sourcePortIndex;
			int targetPortIndex;
			int ea = 0;
			int p;

			// read from String
			for (final String element : edgeInfo) {

				// get edge info of a single source
				final Object[] edgeList = this.edges[ea++];
				final String[] edgeProps = element.split(" ");
				p = 0;

				// get source index and port
				id = edgeProps[p++];
				final List<Object> sourceList = this.children.get(id);
				final KNode sourceNode = (KNode) sourceList.get(0);

				sourcePortIndex = Integer.parseInt(edgeProps[p++]);
				edgeList[2] = sourcePortIndex;

				final boolean hasSourcePort = ++sourcePortIndex != 0;
				final KPort sourcePort;
				if (hasSourcePort) {
					sourcePort = (KPort) sourceList.get(++sourcePortIndex);
				} else {
					sourcePort = null;
				}

				// get target ports
				final int epl = edgeProps.length;
				while (p < epl) {
					// init edge
					final KEdge edge = KimlUtil.createInitializedEdge();
					edgeList[0] = edge;
					edgeList[1] = id;

					// get target index and port
					id = edgeProps[p++];
					edgeList[3] = id;
					final List<Object> targetList = this.children.get(id);

					// select connected nodes
					edge.setSource(sourceNode);
					edge.setTarget((KNode) targetList.get(0));

					targetPortIndex = Integer.parseInt(edgeProps[p++]);
					edgeList[4] = targetPortIndex;

					if (++targetPortIndex != 0) {
						final KPort targetPort = (KPort) targetList.get(++targetPortIndex);

						// set target port of edge
						edge.setTargetPort(targetPort);
						targetPort.getEdges().add(edge);
					}

					if (hasSourcePort) {
						// set source port of edge
						edge.setSourcePort(sourcePort);
						sourcePort.getEdges().add(edge);
					}

				}

			}
		}

		private static String[] correctEmptyArray(final String[] arr) {
			if ((arr.length == 1) && arr[0].isEmpty()) {
				return new String[0];
			} else {
				return arr;
			}
		}
	}
}
