/***************************************************************************
 * Copyright 2013 Kieker Project (http://kieker-monitoring.net)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ***************************************************************************/

package kieker.webgui.service.impl;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.Semaphore;

import javax.annotation.PostConstruct;

import kieker.analysis.AnalysisController.STATE;
import kieker.analysis.model.analysisMetaModel.MIProject;
import kieker.webgui.common.ClassContainer;
import kieker.webgui.common.exception.AnalysisInitializationException;
import kieker.webgui.common.exception.DisplayNotFoundException;
import kieker.webgui.common.exception.InvalidAnalysisStateException;
import kieker.webgui.common.exception.LockProjectException;
import kieker.webgui.common.exception.NewerProjectException;
import kieker.webgui.common.exception.ProjectAlreadyExistingException;
import kieker.webgui.common.exception.ProjectNotExistingException;
import kieker.webgui.common.exception.ProjectStillRunningException;
import kieker.webgui.domain.ComponentListContainer;
import kieker.webgui.domain.DisplayType;
import kieker.webgui.persistence.IProjectDAO;
import kieker.webgui.service.IProjectService;

import org.primefaces.model.UploadedFile;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;

/**
 * This is an implementation of the {@link IProjectService} interface for the project management. It makes sure that accesses to the projects are synchronized. Most
 * of the work is delegated to the underlying layer or to the utility classes.
 * 
 * @author Nils Christian Ehmke
 */
@Service
public final class ProjectServiceImpl implements IProjectService {

	private final List<String> projects = Collections.synchronizedList(new ArrayList<String>());

	private final ConcurrentHashMap<String, Semaphore> fileSystemLocks = new ConcurrentHashMap<String, Semaphore>();
	private final ConcurrentHashMap<String, Semaphore> analysesLocks = new ConcurrentHashMap<String, Semaphore>();

	@Autowired
	private IProjectDAO projectDAO;
	@Autowired
	private AnalysisManagementService acManager;

	/**
	 * Default constructor. <b>Do not use this constructor. This bean is Spring managed.</b>
	 */
	public ProjectServiceImpl() {
		// No code necessary.
	}

	/**
	 * This method does some initialization work after construction. <b>Do not call this method. The method is Spring managed.</b>
	 */
	@PostConstruct
	protected void initialize() {
		// Load a list with all available projects on the FS
		this.projects.addAll(this.projectDAO.listAllProjects());
	}

	@Override
	public void addProject(final String projectName, final String username) throws ProjectAlreadyExistingException, IOException, LockProjectException {
		ProjectServiceImpl.tryLockProject(projectName, this.fileSystemLocks);

		try {
			this.projectDAO.addProject(projectName, username);
			this.projects.add(projectName);
		} finally {
			ProjectServiceImpl.unlockProject(projectName, this.fileSystemLocks);
		}
	}

	@Override
	public void deleteProject(final String projectName) throws ProjectNotExistingException, IOException, ProjectStillRunningException, LockProjectException {
		ProjectServiceImpl.tryLockProjects(projectName, this.fileSystemLocks, projectName, this.analysesLocks);

		final Semaphore oldLockObjectA = this.analysesLocks.get(projectName);
		final Semaphore oldLockObjectB = this.fileSystemLocks.get(projectName);
		try {
			final STATE state = this.acManager.getCurrentState(projectName);
			if ((state == STATE.RUNNING) || (state == STATE.TERMINATING)) {
				throw new ProjectStillRunningException("The project is still running and cannot be deleted.");
			}
			this.projectDAO.removeProject(projectName);
			this.projects.remove(projectName);

			// Remove the old lock objects
			this.analysesLocks.remove(projectName);
			this.fileSystemLocks.remove(projectName);
		} finally {
			oldLockObjectA.release();
			oldLockObjectB.release();
		}
	}

	@Override
	public void copyProject(final String originalProjectName, final String newProjectName) throws ProjectNotExistingException, ProjectAlreadyExistingException,
			IOException, LockProjectException {
		if (originalProjectName.equals(newProjectName)) {
			throw new ProjectAlreadyExistingException("Source and target are identical");
		}
		ProjectServiceImpl.tryLockProjects(originalProjectName, this.fileSystemLocks, newProjectName, this.fileSystemLocks);

		try {
			this.projectDAO.copyProject(originalProjectName, newProjectName);
			this.projects.add(newProjectName);
		} finally {
			ProjectServiceImpl.unlockProject(originalProjectName, this.fileSystemLocks);
			ProjectServiceImpl.unlockProject(newProjectName, this.fileSystemLocks);
		}
	}

	@Override
	public void renameProject(final String originalProjectName, final String newProjectName) throws ProjectNotExistingException, ProjectAlreadyExistingException,
			IOException, LockProjectException {
		if (originalProjectName.equals(newProjectName)) {
			throw new ProjectAlreadyExistingException("Source and target are identical");
		}
		ProjectServiceImpl.tryLockProjects(originalProjectName, this.fileSystemLocks, newProjectName, this.fileSystemLocks);

		try {
			this.projectDAO.copyProject(originalProjectName, newProjectName);
			this.projectDAO.removeProject(originalProjectName);
			synchronized (this.projects) {
				this.projects.add(newProjectName);
				this.projects.remove(originalProjectName);
			}
		} finally {
			ProjectServiceImpl.unlockProject(originalProjectName, this.fileSystemLocks);
			ProjectServiceImpl.unlockProject(newProjectName, this.fileSystemLocks);
		}
	}

	@Override
	public void importProject(final String projectName, final String username, final UploadedFile file) throws ProjectAlreadyExistingException, IOException,
			LockProjectException {
		ProjectServiceImpl.tryLockProject(projectName, this.fileSystemLocks);

		try {
			this.projectDAO.importProject(projectName, username, file);
			this.projects.add(projectName);
		} finally {
			ProjectServiceImpl.unlockProject(projectName, this.fileSystemLocks);
		}
	}

	@Override
	public MIProject loadProject(final String projectName) throws ProjectNotExistingException, IOException, LockProjectException {
		ProjectServiceImpl.tryLockProject(projectName, this.fileSystemLocks);

		try {
			return this.projectDAO.loadProject(projectName);
		} finally {
			ProjectServiceImpl.unlockProject(projectName, this.fileSystemLocks);
		}
	}

	@Override
	public Object loadProject(final String projectName, final ClassContainer classAndMethodContainer) throws ProjectNotExistingException, IOException,
			LockProjectException {
		ProjectServiceImpl.tryLockProject(projectName, this.fileSystemLocks);

		try {
			return this.projectDAO.loadProject(projectName, classAndMethodContainer);
		} finally {
			ProjectServiceImpl.unlockProject(projectName, this.fileSystemLocks);
		}
	}

	@Override
	public void saveProject(final String projectName, final MIProject project, final long timeStamp, final boolean overwriteNewerProject, final String username,
			final String analysisLayout, final String cockpitLayout) throws ProjectNotExistingException, IOException, NewerProjectException, LockProjectException {
		ProjectServiceImpl.tryLockProject(projectName, this.fileSystemLocks);

		try {
			this.projectDAO.saveProject(projectName, project, timeStamp, overwriteNewerProject, username, analysisLayout, cockpitLayout);
		} finally {
			ProjectServiceImpl.unlockProject(projectName, this.fileSystemLocks);
		}
	}

	@Override
	public long getCurrTimeStamp(final String projectName) throws ProjectNotExistingException, LockProjectException {
		ProjectServiceImpl.tryLockProject(projectName, this.fileSystemLocks);

		try {
			return this.projectDAO.getCurrTimeStamp(projectName);
		} finally {
			ProjectServiceImpl.unlockProject(projectName, this.fileSystemLocks);
		}
	}

	@Override
	public void uploadLibrary(final UploadedFile file, final String projectName) throws ProjectNotExistingException, IOException, LockProjectException {
		ProjectServiceImpl.tryLockProject(projectName, this.fileSystemLocks);

		try {
			this.projectDAO.uploadLibrary(file, projectName);
		} finally {
			ProjectServiceImpl.unlockProject(projectName, this.fileSystemLocks);
		}
	}

	@Override
	public ClassLoader getClassLoader(final String projectName, final Object requester) throws ProjectNotExistingException, IOException, LockProjectException {
		ProjectServiceImpl.tryLockProject(projectName, this.fileSystemLocks);

		try {
			return this.projectDAO.getClassLoader(projectName, requester); // NOPMD (ClassLoader)
		} finally {
			ProjectServiceImpl.unlockProject(projectName, this.fileSystemLocks);
		}
	}

	@Override
	public ComponentListContainer getAvailableComponents(final String projectName) throws LockProjectException {
		ProjectServiceImpl.tryLockProject(projectName, this.fileSystemLocks);

		try {
			return this.projectDAO.getAvailableComponents(projectName);
		} finally {
			ProjectServiceImpl.unlockProject(projectName, this.fileSystemLocks);
		}
	}

	@Override
	public List<String> listAllLibraries(final String projectName) throws ProjectNotExistingException, LockProjectException {
		ProjectServiceImpl.tryLockProject(projectName, this.fileSystemLocks);

		try {
			return this.projectDAO.listAllLibraries(projectName);
		} finally {
			ProjectServiceImpl.unlockProject(projectName, this.fileSystemLocks);
		}
	}

	@Override
	public void initializeAnalysis(final String projectName, final ClassLoader classLoader) throws ProjectNotExistingException, InvalidAnalysisStateException,
			AnalysisInitializationException, LockProjectException {
		// We have to lock both - the project and the analysis, as a file has to be loaded
		ProjectServiceImpl.tryLockProjects(projectName, this.fileSystemLocks, projectName, this.analysesLocks);

		try {
			this.acManager.initializeAnalysis(projectName, classLoader);
		} finally {
			ProjectServiceImpl.unlockProject(projectName, this.analysesLocks);
			ProjectServiceImpl.unlockProject(projectName, this.fileSystemLocks);
		}
	}

	@Override
	public void cleanAnalysis(final String projectName) throws ProjectNotExistingException, InvalidAnalysisStateException, LockProjectException {
		ProjectServiceImpl.tryLockProject(projectName, this.analysesLocks);

		try {
			this.acManager.cleanAnalysis(projectName);
		} finally {
			ProjectServiceImpl.unlockProject(projectName, this.analysesLocks);
		}
	}

	@Override
	public void startAnalysis(final String projectName) throws ProjectNotExistingException, InvalidAnalysisStateException, LockProjectException {
		ProjectServiceImpl.tryLockProject(projectName, this.analysesLocks);

		try {
			this.acManager.startAnalysis(projectName);
		} finally {
			ProjectServiceImpl.unlockProject(projectName, this.analysesLocks);
		}
	}

	@Override
	public void stopAnalysis(final String projectName) throws ProjectNotExistingException, InvalidAnalysisStateException, LockProjectException {
		ProjectServiceImpl.tryLockProject(projectName, this.analysesLocks);

		try {
			this.acManager.stopAnalysis(projectName);
		} finally {
			ProjectServiceImpl.unlockProject(projectName, this.analysesLocks);
		}
	}

	@Override
	public Object getDisplay(final String projectName, final String viewName, final String displayName) throws InvalidAnalysisStateException,
			DisplayNotFoundException, LockProjectException {
		ProjectServiceImpl.tryLockProject(projectName, this.analysesLocks);

		try {
			return this.acManager.getDisplay(projectName, viewName, displayName);
		} finally {
			ProjectServiceImpl.unlockProject(projectName, this.analysesLocks);
		}
	}

	@Override
	public STATE getCurrentState(final String projectName) throws LockProjectException {
		ProjectServiceImpl.tryLockProject(projectName, this.analysesLocks);

		try {
			return this.acManager.getCurrentState(projectName);
		} finally {
			ProjectServiceImpl.unlockProject(projectName, this.analysesLocks);
		}
	}

	@Override
	public Object[] getLogEntries(final String projectName) throws InvalidAnalysisStateException, LockProjectException {
		ProjectServiceImpl.tryLockProject(projectName, this.analysesLocks);

		try {
			return this.acManager.getLogEntries(projectName);
		} finally {
			ProjectServiceImpl.unlockProject(projectName, this.analysesLocks);
		}
	}

	@Override
	public boolean deleteLibrary(final String projectName, final String libName) throws IOException, ProjectNotExistingException, LockProjectException {
		ProjectServiceImpl.tryLockProject(projectName, this.fileSystemLocks);

		try {
			return this.projectDAO.deleteLibrary(projectName, libName);
		} finally {
			ProjectServiceImpl.unlockProject(projectName, this.fileSystemLocks);
		}
	}

	@Override
	public String getOwner(final String projectName) throws ProjectNotExistingException, LockProjectException {
		ProjectServiceImpl.tryLockProject(projectName, this.fileSystemLocks);

		try {
			return this.projectDAO.getOwner(projectName);
		} finally {
			ProjectServiceImpl.unlockProject(projectName, this.fileSystemLocks);
		}
	}

	@Override
	public String getLastUser(final String projectName) throws ProjectNotExistingException, LockProjectException {
		ProjectServiceImpl.tryLockProject(projectName, this.fileSystemLocks);

		try {
			return this.projectDAO.getLastUser(projectName);
		} finally {
			ProjectServiceImpl.unlockProject(projectName, this.fileSystemLocks);
		}
	}

	@Override
	public String getAnalysisLayout(final String projectName) throws LockProjectException {
		ProjectServiceImpl.tryLockProject(projectName, this.fileSystemLocks);

		try {
			return this.projectDAO.getAnalysisLayout(projectName);
		} finally {
			ProjectServiceImpl.unlockProject(projectName, this.fileSystemLocks);
		}
	}

	@Override
	public String getCockpitLayout(final String projectName) throws LockProjectException {
		ProjectServiceImpl.tryLockProject(projectName, this.fileSystemLocks);

		try {
			return this.projectDAO.getCockpitLayout(projectName);
		} finally {
			ProjectServiceImpl.unlockProject(projectName, this.fileSystemLocks);
		}
	}

	@Override
	public DisplayType getDisplayType(final String projectName, final String viewName, final String displayConnectorName) throws LockProjectException {
		ProjectServiceImpl.tryLockProject(projectName, this.analysesLocks);

		try {
			return this.acManager.getDisplayType(projectName, viewName, displayConnectorName);
		} finally {
			ProjectServiceImpl.unlockProject(projectName, this.analysesLocks);
		}
	}

	@Override
	public List<String> getProjects() {
		synchronized (this.projects) {
			return new ArrayList<String>(this.projects);
		}
	}

	/**
	 * This method tries to lock the given projects using the given maps. The method makes sure that the projects are locked in a valid order (For this both strings
	 * are compared. If both are equal, the first map will be used for the first lock).
	 * 
	 * @param projectNameA
	 *            The first project to lock.
	 * @param lockMapA
	 *            The map containing the locks.
	 *            * @param projectNameB
	 *            The second project to lock.
	 * @param lockMapB
	 *            The map containing the locks.
	 * 
	 * @throws LockProjectException
	 *             If the project could not be locked.
	 */
	public static void tryLockProjects(final String projectNameA, final ConcurrentHashMap<String, Semaphore> lockMapA, final String projectNameB,
			final ConcurrentHashMap<String, Semaphore> lockMapB) throws LockProjectException {
		final String firstProject;
		final String secondProject;
		final ConcurrentHashMap<String, Semaphore> firstLockMap;
		final ConcurrentHashMap<String, Semaphore> secondLockMap;

		// Find the correct order
		final int compareValue = projectNameA.compareTo(projectNameB);
		if (compareValue >= 0) {
			firstProject = projectNameA;
			secondProject = projectNameB;
			firstLockMap = lockMapA;
			secondLockMap = lockMapB;
		} else {
			firstProject = projectNameB;
			secondProject = projectNameA;
			firstLockMap = lockMapB;
			secondLockMap = lockMapA;
		}

		// Now try to lock both projects. If the locking of the second project fails, we have to unlock the first one though
		ProjectServiceImpl.tryLockProject(firstProject, firstLockMap);

		try {
			ProjectServiceImpl.tryLockProject(secondProject, secondLockMap);
		} catch (final LockProjectException ex) {
			ProjectServiceImpl.unlockProject(firstProject, firstLockMap);
			throw ex;
		}
	}

	/**
	 * This method tries to lock the given project using the given map.
	 * 
	 * @param projectName
	 *            The project to lock.
	 * @param lockMap
	 *            The map containing the locks.
	 * 
	 * @throws LockProjectException
	 *             If the project could not be locked.
	 */
	private static void tryLockProject(final String projectName, final ConcurrentHashMap<String, Semaphore> lockMap) throws LockProjectException {
		// As we also remove locks from time to time, we have to check whether the acquired lock is valid. If we fail to acquire the correct lock too often, the
		// locking failed.
		int failCounter = 0;

		while (true) {
			final Semaphore projectLock = ProjectServiceImpl.getLock(projectName, lockMap);
			projectLock.acquireUninterruptibly();

			if (lockMap.get(projectName) == projectLock) {
				// Everything went well
				return;
			} else {
				// The lock is invalid
				projectLock.release();
				failCounter++;

				if (failCounter > 10) {
					// Abort
					throw new LockProjectException("The project could not be locked.");
				}
			}
		}
	}

	/**
	 * Unlocks the given project.
	 * 
	 * @param projectName
	 *            The project to unlock.
	 * @param lockMap
	 *            The map containing the locks.
	 */
	private static void unlockProject(final String projectName, final ConcurrentHashMap<String, Semaphore> lockMap) {
		lockMap.get(projectName).release();
	}

	/**
	 * This method can be used to get the lock for a given project. If the lock doesn't exist already, a new one will be created.
	 * 
	 * @param projectName
	 *            The project whose lock should be delivered.
	 * @param lockMap
	 *            The map containing the locks.
	 * 
	 * @return A lock object for the project.
	 */
	private static Semaphore getLock(final String projectName, final ConcurrentHashMap<String, Semaphore> lockMap) {
		// Create a new lock and put it into the map - if it doesn't already contain the key
		final Semaphore newLock = new Semaphore(1);
		final Semaphore existLock = lockMap.putIfAbsent(projectName, newLock);

		// Find out which of the both locks is the correct one
		final Semaphore lock;
		if (existLock != null) {
			lock = existLock;
		} else {
			lock = newLock;
		}

		return lock;
	}

}
